"""This module is about compressing the paths of different lengths (either taking into account the sign of connections or not), and looking at the result (`result_summary()`).
"""

import torch
from tqdm import tqdm
import pandas as pd
import plotly.express as px
from scipy.sparse import issparse
import ipywidgets as widgets
import seaborn as sns
import matplotlib.pyplot as plt
from IPython.display import display

from .utils import dynamic_representation, torch_sparse_where, to_nparray, tensor_to_csc


def compress_paths(inprop, step_number, threshold=0, output_threshold=1e-4):
    """
    Performs iterative multiplication of a sparse matrix `inprop` for a specified number of steps, 
    applying thresholding to filter out values below a certain `threshold` to optimize memory usage 
    and computation speed. The function is optimized to run on GPU if available. 

    This function multiplies the connectivity matrix (input in rows; output in columns) `inprop` with itself `step_number` times, 
    with each step's result being thresholded to zero out elements below a given threshold. 
    The function stores each step's result in a list, where each result is further 
    processed to drop values below the `output_threshold` to save memory.

    Args:
        inprop (scipy.sparse.matrix): The connectivity matrix as a scipy sparse matrix.
        step_number (int): The number of iterations to perform the matrix multiplication.
        threshold (float, optional): The threshold value to apply after each multiplication. 
                                     Values below this threshold are set to zero. Defaults to 0.
        output_threshold (float, optional): The threshold value to apply to the final output, 
                                            used to reduce memory footprint. Defaults to 1e-4.

    Returns:
        list: A list of scipy.sparse.csc_matrix objects, each representing connectivity from all neurons to all neurons n steps away. 

    Note:
        This function requires PyTorch and is designed to automatically utilize CUDA-enabled GPU devices 
        if available to accelerate computations. The input matrix `inprop` is converted to a dense tensor 
        before processing. 

    Example:
        >>> from scipy.sparse import csr_matrix
        >>> import numpy as np
        >>> inprop = csr_matrix(np.array([[0.1, 0.2], [0.3, 0.4]]))
        >>> step_number = 2
        >>> compressed_paths = compress_paths(inprop, step_number, threshold=0.1, output_threshold=0.01)
        >>> print(compressed_paths)
    """
    steps_fast = []
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    inprop_tensor = torch.tensor(inprop.toarray(), device=device)

    for i in tqdm(range(step_number)):
        if i == 0:
            out_tensor = inprop_tensor
        else:
            out_tensor = torch.matmul(out_tensor, inprop_tensor)

            # # Downgrade the previous one to save memory
            # steps_fast[-1] = torch.where(steps_fast[-1] >= output_threshold,
            #                              steps_fast[-1], torch.tensor(0.0, device=device))
            # steps_fast[-1] = steps_fast[-1].to_sparse()

        # Thresholding during matmul
        if threshold != 0:
            out_tensor = torch.where(
                out_tensor >= threshold, out_tensor, torch.tensor(0.0, device=device))

        # Thresholding for output
        out_csc = torch.where(out_tensor >= output_threshold,
                              out_tensor, torch.tensor(0.0, device=device))
        out_csc = tensor_to_csc(out_csc.to('cpu'))
        steps_fast.append(out_csc)

    torch.cuda.empty_cache()
    return steps_fast


def compress_paths_signed(inprop, idx_to_sign, target_layer_number, threshold=0, output_threshold=1e-4):
    """
    Calculates the excitatory and inhibitory influences across specified layers of a neural network, using PyTorch for GPU acceleration. This function processes a connectivity matrix (where presynaptic neurons are represented by rows and postsynaptic neurons by columns) to distinguish and compute the influence of excitatory and inhibitory neurons at each layer.

    Args:
        inprop (scipy.sparse.csc_matrix): The initial connectivity matrix representing direct connections between adjacent layers. 
        idx_to_sign (dict): A dictionary mapping neuron indices to their types (1 for excitatory, -1 for inhibitory), used to differentiate between excitatory and inhibitory influences.
        target_layer_number (int): The number of layers through which to calculate influences, starting from the second layer (with the first layer's influence, the direct connectivity, being defined by `inprop`).
        threshold (float, optional): A value to threshold the influences; influences below this value are set to zero, and not passed on in future layers. Defaults to 0.
        output_threshold (float, optional): A threshold for the final output to reduce memory usage, with values below this threshold set to zero. Defaults to 1e-4.

    Returns:
        Tuple[List[scipy.sparse.csc_matrix], List[scipy.sparse.csc_matrix]]: Two lists of sparse matrices representing the excitatory and inhibitory influences, respectively, up to the specified target layer. 

    Note:
        This function is ideal with GPU support. Ensure your environment supports CUDA and that PyTorch is correctly installed.
    """

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    inprop_tensor = torch.tensor(
        inprop.toarray(), device=device, dtype=torch.float32)

    # Create masks for excitatory and inhibitory neurons
    n_neurons = inprop_tensor.shape[0]
    excitatory_mask = torch.tensor([1 if idx_to_sign[i] == 1 else 0 for i in range(
        n_neurons)], dtype=torch.float32, device=device)
    inhibitory_mask = torch.tensor(
        [1 if idx_to_sign[i] == -1 else 0 for i in range(n_neurons)], dtype=torch.float32, device=device)
    print('Number of excitatory neurons: ' +
          str(excitatory_mask.unique(return_counts=True)[1][1].cpu().numpy()))
    print('Number of inhibitory neurons: ' +
          str(inhibitory_mask.unique(return_counts=True)[1][1].cpu().numpy()))

    lne = torch.matmul(torch.diag(excitatory_mask), inprop_tensor)
    lni = torch.matmul(torch.diag(inhibitory_mask), inprop_tensor)

    steps_excitatory = []
    steps_inhibitory = []

    for layer in tqdm(range(target_layer_number)):
        if layer != 0:
          # if layer ==0, return the direct excitatory and inhibitory connections separately
            lne_new = torch.matmul(
                lne, steps_excitatory[0]) + torch.matmul(lni, steps_inhibitory[0])
            lni_new = torch.matmul(
                lne, steps_inhibitory[0]) + torch.matmul(lni, steps_excitatory[0])

            # Apply thresholding
            if threshold > 0:
                lne_new = torch_sparse_where(lne_new, threshold)
                lni_new = torch_sparse_where(lni_new, threshold)

            # Dynamic representation based on density
            lne = dynamic_representation(lne_new)
            lni = dynamic_representation(lni_new)
            torch.cuda.empty_cache()

        # steps_excitatory.append(lne)
        # steps_inhibitory.append(lni)

        # Thresholding for output
        stepe_csc = torch_sparse_where(lne, output_threshold)
        stepe_csc = tensor_to_csc(stepe_csc.to('cpu'))
        steps_excitatory.append(stepe_csc)

        stepi_csc = torch_sparse_where(lni, output_threshold)
        stepi_csc = tensor_to_csc(stepi_csc.to('cpu'))
        steps_inhibitory.append(stepi_csc)

    #     # Downgrade previous matrices to save memory
    #     if steps_excitatory:
    #         steps_excitatory[-1] = torch_sparse_where(
    #             steps_excitatory[-1], output_threshold).to_sparse()
    #         steps_inhibitory[-1] = torch_sparse_where(
    #             steps_inhibitory[-1], output_threshold).to_sparse()

    # # Downgrade the last one to save memory
    # steps_excitatory[-1] = torch_sparse_where(
    #     steps_excitatory[-1], output_threshold).to_sparse()
    # steps_inhibitory[-1] = torch_sparse_where(
    #     steps_inhibitory[-1], output_threshold).to_sparse()
    torch.cuda.empty_cache()

    return steps_excitatory, steps_inhibitory


def add_first_n_matrices(matrices, n):
    """
    Adds the first N connectivity matrices from a list, supporting different path lengths. This function is designed to work with scipy sparse matrices, and dense numpy matrices. Each matrix in the list represents connectivity information for a specific path length.

    Args:
        matrices (list): A list of connectivity matrices of different path lengths. The matrices can be scipy sparse matrices, or dense numpy arrays. The function expects all matrices in the list to be of the same type and shape.
        n (int): The number of initial matrices in the list to be summed. This number must not exceed the length of the matrices list.

    Returns:
        matrix: The resulting matrix after summing the first N matrices. The type of the returned matrix matches the type of the input matrices (scipy sparse matrix, or numpy array).

    Raises:
        ValueError: If the list of matrices is empty or if n is larger than the number of matrices available in the list.

    Example:
        >>> from scipy.sparse import csr_matrix
        >>> matrices = [csr_matrix([[1, 2], [3, 4]]), csr_matrix([[5, 6], [7, 8]]), csr_matrix([[9, 10], [11, 12]])]
        >>> n = 2
        >>> result_matrix = add_first_n_matrices(matrices, n)
        >>> print(result_matrix.toarray())
        [[ 6  8]
        [10 12]]

    Note:
        Ensure that all matrices in the list are of compatible types and shapes before using this function.
    """

    if not matrices:
        raise ValueError("The list of matrices is empty")
    if n > len(matrices):
        raise ValueError("n is larger than the number of matrices available")

    sum_matrix = matrices[0].copy()
    for i in range(1, n):
        sum_matrix += matrices[i]

    return sum_matrix


def result_summary(stepsn, inidx, outidx, inidx_map=None, outidx_map=None, display_output=True, sort_by_column=None, pre_in_column=False, display_threshold=0, include_undefined_groups=False):
    """
    Generates a summary of connections between different types of neurons, 
    represented by their input and output indexes. The function calculates 
    the total synaptic input from presynaptic neuron groups to an average neuron in each 
    postsynaptic neuron group.

    Args:
        stepsn (scipy.sparse matrix or numpy.ndarray): Matrix representing the synaptic strengths 
            between neurons, can be dense or sparse.
        inidx (int, float, list, set, numpy.ndarray, or pandas.Series): Array of indices representing the input (presynaptic) neurons, used to subset stepsn. nan values are removed.
        outidx (int, float, list, set, numpy.ndarray, or pandas.Series): Array of indices representing the output (postsynaptic) neurons.
        inidx_map (dict, optional): Mapping from indices to neuron groups for the input neurons. Defaults to None, in which case neurons are not grouped. 
        outidx_map (dict, optional): Mapping from indices to neuron groups for the output neurons.
            Defaults to None, in which case it is set to be the same as inidx_map.
        display_output (bool, optional): Whether to display the output in a coloured dataframe. Defaults to True.
        sort_by_column (str or list, optional): the column name(s) to sort the result by. If none is provided, then sort by the first column. 
        pre_in_column (bool, optional): Whether to have the presynaptic neuron groups as columns. Defaults to False (pre in rows, post in columns).
        display_threshold (float, optional): The threshold for displaying the output. Defaults to 0.
        include_undefined_groups (bool, optional): Whether to include undefined groups in the output. Defaults to False.

    Returns:
        pd.DataFrame: A dataframe representing the summed synaptic input from presynaptic neuron groups 
            to an average neuron in each postsynaptic neuron group. This dataframe is always returned, regardless of the
            value of display_output.

    Displays:
        If display_output is True, the function will display a styled version of the resulting dataframe.
    """
    if inidx_map is None:
        inidx_map = {idx: idx for idx in inidx}
    if outidx_map is None:
        outidx_map = inidx_map

    # remove nan values in inidx and outidx
    inidx = to_nparray(inidx)
    outidx = to_nparray(outidx)

    if issparse(stepsn):
        matrix = stepsn[:, outidx][inidx, :].toarray()
    else:
        matrix = stepsn[inidx, :][:, outidx]

    if include_undefined_groups:
        # fill the nan values in inidx_map (e.g. 17726: nan) and outidx_map with 'undefined'
        inidx_map = {k: v if pd.notna(
            v) else 'undefined' for k, v in inidx_map.items()}
        outidx_map = {k: v if pd.notna(
            v) else 'undefined' for k, v in outidx_map.items()}

    # Create the dataframe
    df = pd.DataFrame(data=matrix,
                      # choose what to group by here
                      # if idx is mapped to root_id, if root_id is kept as int64, the root_ids seem a bit messed up
                      index=[str(inidx_map[key]) for key in inidx],
                      columns=[str(outidx_map[key]) for key in outidx])

    # Sum across rows: presynaptic neuron is in the rows
    # summing across neurons of the same type: total amount of input from that type for the postsynaptic neuron
    summed_df = df.groupby(df.index).sum()

    # Average across columns and transpose back
    # averaging across columns of the same type:
    # on average, a neuron of that type receives x% input from a presynaptic type
    result_df = summed_df.T.groupby(level=0).mean().T

    if pre_in_column:
        result_df = result_df.T

    if display_threshold > 0:
        # only display rows where any value >= display_threshold
        result_df = result_df[(result_df >= display_threshold).any(axis=1)]

    if sort_by_column is None:
        # sort result_df by the values in the first column, in descending order
        result_df = result_df.sort_values(
            by=result_df.columns[0], ascending=False)
    elif isinstance(sort_by_column, str):
        sort_by_column = [sort_by_column]

    if sort_by_column is not None:
        if set(sort_by_column).issubset(result_df.columns):
            result_df = result_df.sort_values(
                by=sort_by_column, ascending=False)
        else:
            raise ValueError(
                "sort_by_column must be present in the values of outidx_map.")

    if display_output:
        result_dp = result_df.style.background_gradient(cmap='Blues', vmin=result_df.min().min(),
                                                        vmax=result_df.max().max())
        display(result_dp)
    return result_df


def contribution_by_path_lengths(steps, inidx, outidx, outidx_map):
    """
    Analyzes the contribution of presynaptic neurons to postsynaptic neuron groups across 
    different path lengths. This function calculates and visualizes the average input 
    received by a neuron in each postsynaptic neuron group from presynaptic ones, aggregated over specified 
    path lengths. Direct connections are in path_length 1.

    Args:
        steps (list of scipy.sparse matrices): List of sparse matrices, each representing 
            synaptic strengths for a specific path length.
        inidx (int, float, list, set, numpy.ndarray, or pandas.Series): Array of indices representing input (presynaptic) neurons.
        outidx (int, float, list, set, numpy.ndarray, or pandas.Series): Array of indices representing output (postsynaptic) neurons.
        outidx_map (dict): Mapping from indices to postsynaptic neuron groups.

    Returns:
        None: Displays a line plot with path lengths on the x-axis, the average input 
            received on the y-axis, and lines differentiated by postsynaptic neuron groups.
            The plot represents how different postsynaptic neuron groups are influenced by 
            presynaptic neurons over various path lengths.

    The function iterates through each path length (each 'step'), computes the sum of inputs 
    from presynaptic to each postsynaptic neuron. It then calculates the average for each postsynaptic group. It then generates a plot showing the relationship between path lengths and synaptic input 
    contribution for each postsynaptic neuron group.
    """

    # remove nan values in inidx and outidx
    inidx = to_nparray(inidx)
    outidx = to_nparray(outidx)

    rows = []
    for step in steps:
        df = pd.DataFrame(data=step[:, outidx][inidx, :].toarray(),
                          # all input are grouped togehter.
                          columns=[outidx_map[key] for key in outidx])

        # Sum across rows: presynaptic neuron is in the rows
        # summing across neurons of the same type: total amount of input for the postsynaptic neuron
        # this gives a dataframe of one column, where each row is a value from outidx_map
        # there is some secret transposing after summing
        summed_df = df.sum().to_frame()

        # Average across columns and transpose back
        # averaging across columns of the same type:
        # on average, a neuron of that type receives x% input from the inidx
        result_df = summed_df.groupby(level=0).mean().T
        rows.append(result_df)

    # first have a dataframe where: each row is a path length, each column is a postsynaptic cell type
    # then pivot_wide to long
    # index is the path length
    # variable is postynaptic cell type
    # value is y axis
    contri = pd.concat(rows, ignore_index=True).melt(
        ignore_index=False).reset_index()
    contri.columns = ['path_length', 'postsynaptic_type', 'value']
    contri.path_length = contri.path_length+1

    fig = px.line(contri, x="path_length", y="value",
                  color='postsynaptic_type')
    fig.update_layout(
        xaxis=dict(
            tickmode='linear',
            tick0=1,
            dtick=1
        )
    )
    fig.show()


def contribution_by_path_lengths_heatmap(steps, inidx, outidx, inidx_map=None, outidx_map=None,
                                         sort_by_index=True, sort_by_column=None,
                                         pre_in_column=False,
                                         display_threshold=0,
                                         cmap='viridis'):
    """
    Display the contribution from inidx to outidx, grouped by inidx_map and outidx_map, across different path lengths.

    Args:
        steps (list of scipy.sparse matrices): List of sparse matrices, each representing synaptic strengths for a specific path length.
        inidx (int, float, list, set, numpy.ndarray, or pandas.Series): Array of indices representing input (presynaptic) neurons.
        outidx (int, float, list, set, numpy.ndarray, or pandas.Series): Array of indices representing output (postsynaptic) neurons.
        inidx_map (dict, optional): Mapping from indices to input neuron groups. Defaults to None, in which case neurons are not grouped.
        outidx_map (dict, optional): Mapping from indices to output neuron groups. Defaults to None, in which case it is set to be the same as inidx_map.
        sort_by_index (bool, optional): Whether to sort the output by index. Defaults to True.
        sort_by_column (str or list, optional): the column name(s) to sort the result by. If none is provided, then sort by the first column.
        pre_in_column (bool, optional): Whether to have the presynaptic neuron groups as columns. Defaults to False (pre in rows, post in columns).
        display_threshold (float, optional): The threshold for displaying the output. Defaults to 0.
        cmap (str, optional): The colormap to use for the heatmap. Defaults to 'viridis'.

    Returns:
        None: Displays an interactive heatmap showing the contribution from inidx to outidx, grouped by inidx_map and outidx_map, across different path lengths.

    """

    inidx = to_nparray(inidx)
    outidx = to_nparray(outidx)

    if inidx_map is None:
        inidx_map = {idx: idx for idx in inidx}
    if outidx_map is None:
        outidx_map = inidx_map

    def plot_heatmap(index):
        plt.figure(figsize=(30, 15))
        # plt.imshow(heatmaps[index], cmap='viridis', aspect = 'auto')
        # Use seaborn's heatmap function which is a higher-level API for Matplotlib's imshow
        sns.heatmap(heatmaps[index-1], annot=True, fmt=".2f",
                    linewidths=.5, cmap=cmap)

        # Rotate the tick labels for the columns to show them better
        plt.xticks(rotation=90)
        plt.yticks(rotation=0)

        # Show the heatmap
        plt.show()

    heatmaps = []

    for step in tqdm(steps):
        df = result_summary(step, inidx, outidx,
                            inidx_map, outidx_map,
                            display_output=False,
                            sort_by_column=sort_by_column,
                            pre_in_column=pre_in_column,
                            display_threshold=display_threshold)
        if sort_by_index:
            df.sort_index(inplace=True)
        heatmaps.append(df)

    slider = widgets.IntSlider(value=1, min=1, max=len(heatmaps), step=1,
                               description='Path length', continuous_update=True)

    # Link the slider to the plotting function
    display(widgets.interactive(plot_heatmap, index=slider))



import unittest
from connectome_interpreter.utils import modify_coo_matrix
from scipy.sparse import coo_matrix
import pandas as pd
import numpy as np


class TestModifyCooMatrix(unittest.TestCase):

    def setUp(self):
        # Create a small sparse matrix with float data type and values between 0 and 1
        self.example_coo = coo_matrix(np.array([
            [0.1, 0.2, 0],
            [0, 0, 0.3],
            [0.4, 0, 0.5]
        ], dtype=np.float32))  # Ensure dtype is float32

        # DataFrame for batch updates with float values
        self.updates_df = pd.DataFrame({
            'input_idx': [0, 2],
            'output_idx': [2, 1],
            'value': [0.25, 0.75]  # Use values between 0 and 1
        })

    def test_single_update(self):
        # Test updating a single value with a float value between 0 and 1
        result = modify_coo_matrix(
            self.example_coo, input_idx=1, output_idx=2, value=0.99, re_normalize=False)
        self.assertAlmostEqual(result.toarray()[1, 2], 0.99)

    def test_batch_update(self):
        # Test updating using a DataFrame of updates with float values
        result = modify_coo_matrix(
            self.example_coo, updates_df=self.updates_df, re_normalize=False)
        self.assertAlmostEqual(result.toarray()[0, 2], 0.25)
        self.assertAlmostEqual(result.toarray()[2, 1], 0.75)

    def test_renormalize(self):
        # Test the re-normalization functionality with float values
        result = modify_coo_matrix(
            self.example_coo, updates_df=self.updates_df, re_normalize=True)
        updated_cols = self.updates_df['output_idx'].unique()
        # Check if columns sum up to 1 or close to it
        column_sums = np.array(result.sum(axis=0))[0]
        for col_sum in column_sums[updated_cols]:
            # Assuming a little margin for floating-point arithmetic
            self.assertTrue(0.99 <= col_sum <= 1.01)

    def test_handle_zero_colsums(self):
        # Test how the function handles columns with zero sums after update with float values
        zero_updates_df = pd.DataFrame({
            'input_idx': [1, 0],
            'output_idx': [2, 1],
            'value': [0.0, 0.0]  # Use float zero to keep consistent data type
        })
        result = modify_coo_matrix(
            self.example_coo, updates_df=zero_updates_df, re_normalize=True)
        column_sums = np.array(result.sum(axis=0))[0]
        self.assertAlmostEqual(column_sums[1], 0.0)



import random
import warnings

import torch
import pandas as pd
import numpy as np
from scipy.sparse import coo_matrix, csr_matrix, issparse
import matplotlib.colors as mcl
import matplotlib.pyplot as plt
from tqdm import tqdm
import networkx as nx


def dynamic_representation(tensor, density_threshold=0.2):
    """Convert tensor to sparse if density is below threshold, otherwise to dense."""
    nonzero_elements = torch.nonzero(tensor).size(0)

    # to_sparse() doesn't support operations on tensors with more than INT_MAX (2,147,483,647) elements
    if nonzero_elements < 2147483647:
        total_elements = tensor.numel()
        density = nonzero_elements / total_elements

        if density < density_threshold:
            return tensor.to_sparse()
        else:
            return tensor.to_dense()
    else:
        print("Tensor is too large to convert to sparse. Returning dense tensor.")
        return tensor


def torch_sparse_where(x, threshold):
    """
    Apply a threshold to a tensor, setting values below the threshold to zero. This function allows the tensor to be sparse. torch.where() does not.

    Args:
      x (torch.Tensor): The input tensor to apply the threshold to.
      threshold (float): The threshold value.

    Returns:
      torch.Tensor: A new tensor with values below the threshold set to zero.
    """
    if x.is_sparse:
        values = x._values()
        indices = x._indices()

        # Apply the threshold to the values
        thresholded_values = torch.where(
            values >= threshold, values, torch.tensor(0.0, device=x.device, dtype=x.dtype))

        # Create a new sparse tensor with the same indices but updated values
        result = torch.sparse_coo_tensor(
            indices, thresholded_values, x.size(), device=x.device, dtype=x.dtype).coalesce()
        # It might be beneficial to remove zero values that were newly created by thresholding
        # .coalesce() combines duplicates and removes zeros if present
    else:
        result = torch.where(
            x >= threshold, x, torch.tensor(0.0, device=x.device))

    return result


def tensor_to_csc(tensor):
    """Turn torch.Tensor into scipy Compressed Sparse Column matrix.

    Args:
      tensor (torch.Tensor): A (sparse) tensor.

    Returns:
      scipy.sparse.csc_matrix: A Scipy sparse Compressed Sparse Column matrix.
    """
    if tensor.is_sparse:
        tensor = tensor.coalesce()
        indices = tensor.indices().numpy()
        values = tensor.values().numpy()
        shape = tensor.shape
        coo = coo_matrix((values, (indices[0], indices[1])), shape=shape)
    else:
        # Convert dense tensor to COO directly, necessary if the tensor is not sparse
        indices = torch.nonzero(tensor).t()
        values = tensor[indices[0], indices[1]]
        shape = tensor.shape
        coo = coo_matrix(
            (values.numpy(), (indices[0].numpy(), indices[1].numpy())), shape=shape)

    return coo.tocsc()


def coo_tensor_to_el(coo_tensor):
    """
    Convert a PyTorch sparse COO tensor to a DataFrame representing an edge list.

    This function checks if the input tensor is sparse. If not, it converts it to a sparse COO tensor.
    It then extracts the indices and values, and creates a DataFrame with columns 'pre', 'post', and 'weight'.
    Each row in the DataFrame represents an edge in the graph, where 'pre' and 'post' are the nodes connected by the edge,
    and 'weight' is the weight of the edge.

    Args:
      coo_tensor (torch.Tensor): A PyTorch tensor, either already in sparse COO format or dense.

    Returns:
      pd.DataFrame: A DataFrame with columns 'pre', 'post', and 'weight', representing the edge list of the graph.
    """

    if not coo_tensor.is_sparse:
        coo_tensor = coo_tensor.to_sparse_coo()

    # Transpose and convert to numpy array
    indices = coo_tensor.indices().t().cpu().numpy()
    values = coo_tensor.values().cpu().numpy()

    # Split the indices to row and column
    pre, post = indices[:, 0], indices[:, 1]

    edge_list_df = pd.DataFrame(
        {'pre': pre, 'post': post, 'weight': values})
    return edge_list_df


def coo_to_el(coo, row_indices=None, col_indices=None):
    """
    Extracts an edgelist from a COO matrix, optionally using pre-specified row and/or column indices.
    If row_indices or col_indices are None, all rows or columns are considered, respectively.
    Each row in the resulting DataFrame represents an edge in the graph, where 'pre' and 'post' are the nodes connected by the edge,
    and 'value' is the weight of the edge.

    Args:
      coo: A scipy.sparse.coo_matrix instance.
      row_indices: Optional; A list or array of row indices of interest.
      col_indices: Optional; A list or array of column indices of interest.

    Returns:
      pd.DataFrame: A DataFrame with columns 'pre', 'post', and 'value', representing the edge list of the graph.
    """
    # Determine whether to filter on rows/columns based on provided indices
    row_mask = np.full(coo.shape[0], True) if row_indices is None else np.isin(
        coo.row, row_indices)
    col_mask = np.full(coo.shape[1], True) if col_indices is None else np.isin(
        coo.col, col_indices)

    # Combine row and column masks
    mask = row_mask & col_mask

    # Filter rows, cols, and data based on the combined mask
    rows = coo.row[mask]
    cols = coo.col[mask]
    data = coo.data[mask]

    all_el = pd.DataFrame({'pre': rows, 'post': cols, 'value': data})

    return all_el


def adjacency_df_to_el(adjacency, threshold=None):
    """
    Convert a DataFrame representing an adjacency matrix to an edge list.

    This function takes a DataFrame where the index and columns represent nodes in the graph,
    and the values represent the weights of the edges between them. It converts this DataFrame to an edge list,
    where each row represents an edge in the graph, with columns 'pre', 'post', and 'weight'.

    Args:
      adjacency (pd.DataFrame): A DataFrame representing an adjacency matrix.
      threshold (float, optional): If provided, edges with values below this threshold are removed. Default to None.

    Returns:
      pd.DataFrame: A DataFrame with columns 'pre', 'post', and 'weight', representing the edge list of the graph.
    """
    if threshold is not None:
        adjacency = adjacency.where(adjacency >= threshold, 0)

    # Stack the DataFrame to create a long format
    adjacency = adjacency.stack().reset_index()
    adjacency.columns = ['pre', 'post', 'weight']
    adjacency = adjacency[adjacency['weight'] != 0]

    return adjacency


def modify_coo_matrix(sparse_matrix, input_idx=None, output_idx=None, value=None, updates_df=None, re_normalize=True):
    """
    Modify the values of a sparse matrix (either COO or CSR format) at specified indices.

    There are two modes of operation:
    1. Single update mode: where `input_idx`, `output_idx`, and `value` are provided as individual arguments. In this case all combinations of input_idx and output_idx are updated.
    2. Batch update mode: where `updates_df` is provided, a DataFrame with columns ['input_idx', 'output_idx', 'value'].

    Args:
      sparse_matrix (coo_matrix or csr_matrix): The sparse matrix to modify.
      input_idx (int, list, numpy.ndarray, set, optional): Row indices for updates.
      output_idx (int, list, numpy.ndarray, set, optional): Column indices for updates.
      value (numeric, list, numpy.ndarray, optional): New values for updates. If it is a number, then it is used for all updates. Else, it needs to be of length equal to the product of the lengths of input_idx and output_idx.
      updates_df (DataFrame, optional): The DataFrame containing batch updates.
      re_normalize (bool, optional): Whether to re-normalize the matrix after updating, such that all values in each column sum up to 1. Default to True.

    Note:
      In single update mode, `input_idx`, `output_idx`, and `value` must all be provided.
      In batch update mode, only `updates_df` should be provided.
      If both modes are provided, the function will prioritize the single update mode.

    Returns:
      coo_matrix or csr_matrix: The updated sparse matrix, in the same format as the input.
    """
    if not issparse(sparse_matrix):
        raise TypeError("The provided matrix is not a sparse matrix.")

    # Convert sparse matrix to COO format for direct edge manipulation
    original_format = sparse_matrix.getformat()
    if original_format != 'coo':
        sparse_matrix = sparse_matrix.tocoo()

    # Extract edges as a DataFrame
    edges = pd.DataFrame(
        {'input_idx': sparse_matrix.row, 'output_idx': sparse_matrix.col, 'value': sparse_matrix.data})

    print("Updating matrix...")
    if input_idx is not None and output_idx is not None and value is not None:
        # Create updates DataFrame from provided indices and values
        input_idx = to_nparray(input_idx)
        output_idx = to_nparray(output_idx)
        if np.isscalar(value):
            value = np.full(len(input_idx) * len(output_idx), value)
        elif len(value) != len(input_idx) * len(output_idx):
            raise ValueError(
                "Length of 'value' must match the product of the lengths of 'input_idx' and 'output_idx', or be one single value.")

        updates_df = pd.DataFrame({
            'input_idx': np.repeat(input_idx, len(output_idx)),
            'output_idx': np.tile(output_idx, len(input_idx)),
            'value': value
        })

    elif updates_df is not None:
        # check column names
        if not all(col in updates_df.columns for col in ['input_idx', 'output_idx', 'value']):
            raise ValueError(
                "The DataFrame must contain columns ['input_idx', 'output_idx', 'value'].")

    else:
        raise ValueError(
            "Either ['input_idx', 'output_idx', and 'value'], or 'updates_df' must be provided.")

    # Concatenate edges and updates_df, ensuring that updates_df is last so its values take precedence
    combined_edges = pd.concat([edges, updates_df])

    # Remove duplicates, keeping the last occurrence (from updates_df)
    # We specify the subset to be the columns 'input_idx' and 'output_idx'
    edges = combined_edges.drop_duplicates(
        subset=['input_idx', 'output_idx'], keep='last')

    if re_normalize:
        print("Re-normalizing updated columns...")
        edge_to_normalise = edges[edges.output_idx.isin(updates_df.output_idx)]
        col_sums = edge_to_normalise.groupby('output_idx').value.sum()
        col_sums[col_sums == 0] = 1

        edge_to_normalise.loc[:, ['value']] = edge_to_normalise['value'] / \
            edge_to_normalise['output_idx'].map(col_sums)

        edges = pd.concat([edges[~edges.output_idx.isin(updates_df.output_idx)],
                           edge_to_normalise])

    # Convert back to original sparse matrix format
    updated_matrix = coo_matrix(
        (edges['value'], (edges['input_idx'], edges['output_idx'])), shape=sparse_matrix.shape)

    # if re_normalize:
    #     print("Re-normalizing updated columns...")
    #     # Convert updated_matrix back to CSR format for efficient column operations
    #     updated_matrix = updated_matrix.tocsr()

    #     # Identify the unique columns to re-normalize based on the updates_df
    #     unique_cols = updates_df.reset_index()['output_idx'].unique()

    #     # Compute sum of each column that has been updated
    #     col_sums = np.array(
    #         updated_matrix[:, unique_cols].sum(axis=0)).flatten()

    #     for idx, col in enumerate(unique_cols):
    #         if col_sums[idx] != 0:
    #             updated_matrix[:, col] /= col_sums[idx]

    if issubclass(updated_matrix.dtype.type, np.integer):
        warnings.warn(
            "Matrix data is of integer type instead of float32. Is the sparse matrix normalised connectivity?",
            UserWarning
        )
    updated_matrix.eliminate_zeros()
    return updated_matrix.asformat(original_format)

    # if not issparse(sparse_matrix):
    #     raise TypeError("The provided matrix is not a sparse matrix.")

    # # Ensure we're working with CSR for efficiency, remember original format
    # original_format = sparse_matrix.getformat()
    # csr = sparse_matrix.tocsr() if original_format == 'coo' else sparse_matrix

    # updated_cols = set()  # Track columns that are updated or added
    # print("Updating matrix...")
    # if input_idx is not None and output_idx is not None and value is not None:
    #     # Convert inputs to arrays and ensure compatibility
    #     input_idx = to_nparray(input_idx)
    #     output_idx = to_nparray(output_idx)
    #     if np.isscalar(value) or (isinstance(value, np.ndarray) and value.size == 1):
    #         value = np.full(len(input_idx) * len(output_idx), value)
    #     elif len(value) == len(input_idx) * len(output_idx):
    #         value = np.atleast_1d(value)
    #     else:
    #         raise ValueError(
    #             "The length of 'value' is incorrect. It should either be a single value or match the length of 'input_idx' * 'output_idx'.")

    #     for (i, j, val) in tqdm(zip(np.repeat(input_idx, len(output_idx)),
    #                                 np.tile(output_idx, len(input_idx)),
    #                                 value)):
    #         # Update the matrix, note the changes in indexing for csr
    #         if csr[i, j] != 0:
    #             csr[i, j] = val  # Efficient for CSR format
    #             updated_cols.add(j)

    # elif updates_df is not None:
    #     # Batch updates from a DataFrame
    #     for _, row in tqdm(updates_df.iterrows()):
    #         i, j, val = row['input_idx'], row['output_idx'], row['value']
    #         if csr[i, j] != 0:
    #             csr[i, j] = val  # Efficient for CSR format
    #             updated_cols.add(j)

    # else:
    #     raise ValueError("Either ['input_idx', 'output_idx', and 'value'], or 'updates_df' must be provided.")

    # if re_normalize and updated_cols:
    #     print("Re-normalizing updated columns...")
    #     # Normalize updated columns
    #     for col in updated_cols:
    #         col_sum = csr[:, col].sum()
    #         if col_sum != 0:
    #             csr[:, col] /= col_sum
    # # remove 0 entries
    # csr.eliminate_zeros()
    # return csr.asformat(original_format)


def to_nparray(input_data):
    """
    Converts the input data into a numpy array, filtering out any NaN values and duplicates. 
    The input can be a single number, a list, a set, a numpy array, or a pandas Series.

    Args:
        input_data: The input data to convert. Can be of type int, float, list, set, numpy.ndarray, or pandas.Series.

    Returns:
        numpy.ndarray: A unique numpy array created from the input data, with all NaN values removed.
    """
    # First, ensure the input is in array form and convert pd.Series to np.ndarray
    if isinstance(input_data, (int, float)):
        input_array = np.array([input_data])
    elif isinstance(input_data, (list, set, np.ndarray, pd.Series)):
        input_array = np.array(list(input_data))
    else:
        raise TypeError(
            "Input data must be an int, float, list, set, numpy.ndarray, or pandas.Series")

    # Then, remove NaN values
    cleaned_array = input_array[~pd.isna(input_array)]

    # Finally, return unique values
    return np.unique(cleaned_array)


def get_ngl_link(df, no_connection_invisible=True, colour_saturation=0.4, scene=None, normalise_within_column=True, include_postsynaptic_neuron=False):
    """
    Generates a Neuroglancer link with layers, corresponding to each column in the df. The function
    processes a dataframe, adding colour information to each neuron,
    and then creates layers in a Neuroglancer scene for visualization. Separate layers
    are created for each column in the DataFrame, with unique segment colors based on
    the values in the DataFrame.

    Args:
        df (pandas.DataFrame): A DataFrame containing neuron metadata. The index should contain neuron identifiers (root_ids), and columns should represent different
      attributes or categories.
        no_connection_invisible (bool, optional): Whether to make invisible neurons that are not connected. Default to True (invisible). 
        colour_saturation (float, optional): The saturation of the colours. Default to 0.4.
        scene (ngl.Scene, optional): A Neuroglancer scene object from nglscenes package. You can read a scene from clipboard like `scene = Scene.from_clipboard()`. 
        normalise_within_column (bool, optional): Whether to normalise the values within each column (the alternative is normalising by the min and max value in the entire dataframe). Default to True.
        include_postsynaptic_neuron (bool, optional): Whether to include the postsynaptic neuron (column names of `df`) in the visualisation. Default to False.

    Returns:
        str: A URL to the generated Neuroglancer scene.

    Note:
        The function assumes that the 'scene1' variable is defined in the global scope and is an instance of ngl.Scene.
        The function creates separate Neuroglancer layers for each column in the DataFrame, using the column name as the layer name.
        The root_ids are colored based on the values in the DataFrame, with a color scale ranging from white (minimum value) to a specified color (maximum value).
        The function relies on the Neuroglancer library for layer creation and scene manipulation.
    """

    try:
        import nglscenes as ngl
    except ImportError:
        raise ImportError(
            "To use this function, please install the package by running 'pip3 install git+https://github.com/schlegelp/nglscenes@main'")

    # define a scene if not given:
    if scene == None:
        no_scene_provided = True
        # Initialize a scene
        scene = ngl.Scene()
        scene['layout'] = '3d'
        scene['position'] = [527216.1875, 208847.125, 84774.0625]
        scene['projectionScale'] = 400000
        scene['dimensions'] = {"x": [1e-9, "m"],
                               "y": [1e-9, "m"], "z": [1e-9, "m"]}

        # and another for FAFB mesh
        fafb_layer = ngl.SegmentationLayer(source='precomputed://https://spine.itanna.io/files/eric/jfrc_mesh_test',
                                           name='jfrc_mesh_test1')
        fafb_layer['segments'] = ['1']
        fafb_layer['objectAlpha'] = 0.17
        fafb_layer['selectedAlpha'] = 0.55
        fafb_layer['segmentColors'] = {'1': '#cacdd8'}
        fafb_layer['colorSeed'] = 778769298

        # and the neuropil layer with names
        np_layer = ngl.SegmentationLayer(source='precomputed://gs://neuroglancer-fafb-data/elmr-data/FAFBNP.surf/mesh#type=mesh',
                                         name='neuropil_regions_surface_named')
        np_layer['segments'] = [str(num) for num in range(0, 79)]
        np_layer['visible'] = False
        np_layer['objectAlpha'] = 0.17

    # Define a list of colors optimized for human perception on a dark background
    colors = ['#ff6b6b', '#f06595', '#cc5de8', '#845ef7', '#5c7cfa', '#339af0', '#22b8cf', '#20c997', '#51cf66', '#94d82d', '#fcc419', '#4ecdc4',
              '#ffe66d', '#7bed9f', '#a9def9', '#f694c1', '#c7f0bd', '#ffc5a1', '#ff8c94', '#ffaaa6', '#ffd3b5', '#a8e6cf', '#a6d0e4', '#c1beff', '#f5b0cb']

    # Normalize the values in the DataFrame
    if normalise_within_column:
        df_norm = (df - df.min()) / (df.max() - df.min())
    else:
        df_norm = (df - df.min().min()) / (df.max().max() - df.min().min())

    scene['layout'] = '3d'

    source = ['precomputed://gs://flywire_v141_m783',
              'precomputed://https://flyem.mrc-lmb.cam.ac.uk/flyconnectome/dynann/flytable-info-783-all']

    for i, column in enumerate(df.columns):
        color = random.choice(colors)
        cmap = mcl.LinearSegmentedColormap.from_list(
            "custom_cmap", ["white", color])

        df_group = df_norm[[column]]
        if no_connection_invisible:
            df_group = df_group[df_group.iloc[:, 0] > 0]

        layer = ngl.SegmentationLayer(source=source, name=str(column))

        layer['segments'] = list(df_group.index.astype(str))
        layer['segmentColors'] = {
            # trick from Claud to make the colours more saturated
            str(root_id): mcl.to_hex(cmap(colour_saturation + (1-colour_saturation)*value.values[0]))
            for root_id, value in df_group.iterrows()
        }
        # only the first layer is visible
        if i == 0:
            layer['visible'] = True
        else:
            layer['visible'] = False

        if include_postsynaptic_neuron:
            layer['segments'].append(str(df_group.columns[0]))
            layer['segmentColors'][str(df_group.columns[0])] = '#43A7FF'

        scene.add_layers(layer)

    if no_scene_provided:
        # add FAFB and NP layers at the end
        scene.add_layers(fafb_layer)
        scene.add_layers(np_layer)

    return scene.url


def get_activations(array, global_indices, idx_map=None, top_n=None, threshold=None):
    """
    Get activation for neurons (in rows) in `array` for each time step (in columns). Optionally group neurons by `idx_map`, and filter by `top_n` or `threshold`.

    Args:
        array (np.ndarray): 2D array of neuron activations, where rows represent neurons
            and columns represent different time steps.
        global_indices (int, list, set, np.ndarray, pd.Series): Array of global neuron indices corresponding to keys in `idx_map`. 
        idx_map (dict, optional): Mapping from neuron index (`global_indices`) to neuron identifier. If not None, and if multiple neurons map to the same identifier, the activations are averaged. Defaults to None.
        top_n (int, optional): Number of top activations to return for each column. If None,
            all activations above the threshold are returned. Defaults to None.
        threshold (float, dict, optional): Minimum activation level to consider. If a dictionary is provided, the threshold for each column is specified by the column index. Defaults to None.

    Returns:
        dict: A dictionary where each key is a column index and each value is a nested dictionary
            of neuron identifiers and their activations, for those activations that are either
            in the top n, above the threshold, or both.

    Note:
        The global_indices have to be in the same order as the indices in defining the original model. 
        If both `n` and `threshold` are provided, the function returns up to top n activations
        that are also above the threshold for each column.
    """
    result = {}
    indices = to_nparray(global_indices)
    if array.shape[0] != len(indices):
        raise ValueError(
            "The length of 'global_indices' should match the number of rows in 'array'.")

    global_to_local_map = {global_idx: num for num,
                           global_idx in enumerate(indices)}

    for col in range(array.shape[1]):
        # Determine which indices to use based on the 'sensory' flag
        # these are global indices in the all-to-all connectivity

        column_values = array[:, col]
        local_indices = np.asarray(range(len(column_values)))

        # Filter activations by threshold if provided
        if threshold is not None:
            if isinstance(threshold, dict):
                if col not in threshold:
                    thresh = 0
                else:
                    thresh = threshold[col]
            else:
                thresh = threshold

            if thresh > 0:
                # these are local indices
                local_indices -= np.where(column_values < thresh)[0]

        # Sort the filtered activations
        # these are the local indices
        sorted_indices = np.argsort(
            column_values)[-top_n:] if top_n is not None else np.argsort(column_values)

        # get intersection of sorted_indices and local_indices, in the same order as sorted_indices
        selected_local = [
            index for index in sorted_indices if index in local_indices]
        selected_global = indices[selected_local]

        # Build the result dictionary
        if idx_map is None:
            result[col] = {idx: column_values[global_to_local_map[idx]]
                           for idx in selected_global}
        else:
            # initialise a dict of empty lists
            result[col] = {idx_map[idx]: [] for idx in selected_global}

            for idx in selected_global:
                result[col][idx_map[idx]].append(
                    column_values[global_to_local_map[idx]])
            # grouped indices by idx_map
            new_indices = result[col].keys()
            # calculate the average
            result[col] = {idx: np.mean(result[col][idx])
                           for idx in new_indices}

    return result


def plot_layered_paths(path_df, figsize=(10, 8), priority_indices=None, sort_by_activation=False,
                       fraction=0.03, pad=0.02, weight_decimals=2,
                       neuron_to_sign=None, sign_color_map={1: 'red', -1: 'blue'}):
    """
    Plots a directed graph of layered paths with optional node coloring based on activation values.

    This function creates a visualization of a directed graph with nodes placed in layers. Nodes can be optionally
    colored based on 'pre_activation' and 'post_activation' columns present in the dataframe. If these columns are
    missing, a default color is used for all nodes. The edges are weighted, and their labels represent the weight values.

    Args:
        path_df (pandas.DataFrame): A dataframe containing the columns 'pre', 'post', 'layer', 'weight', and optionally 
                                    'pre_activation', 'post_activation', 'pre_layer', 'post_layer'. Each row represents an
                                    edge in the graph. The 'pre' and 'post' columns refer to the source and target nodes, respectively.
                                    The 'layer' column is used to place nodes in layers, and 'weight' indicates the edge weight.
                                    If present, 'pre_activation' and 'post_activation' are used to color the nodes based on
                                    their activation values.
        figsize (tuple, optional): A tuple indicating the size of the matplotlib figure. Defaults to (10, 8).
        priority_indices (list, optional): A list of indices to prioritize when creating the layered positions. Nodes with these
                                    indices will be placed at the top of their respective layers. Defaults to None. 
        sort_by_activation (bool, optional): A flag to sort the nodes based on their activation values (after grouping by priority). Defaults to False.
        fraction (float, optional): The fraction of the figure width to use for the colorbar. Defaults to 0.03.
        pad (float, optional): The padding between the colorbar and the plot. Defaults to 0.02.
        weight_decimals (int, optional): The number of decimal places to display for edge weights. Defaults to 2.
        neuron_to_sign (dict, optional): A dictionary mapping neuron names (as they appear in path_df) to their signs (e.g. {'KCg-m': 1, 'Delta7': -1}). Defaults to None.
        sign_color_map (dict, optional): A dictionary mapping neuron signs to colors. Defaults to red for excitatory, and blue for inhibitory. Edges are in lightgrey by default.

    Returns:
        None: This function does not return a value. It generates a plot using matplotlib.

    Note:
        If 'pre_layer' and 'post_layer' columns are not in the dataframe, they will be created within the function
        to uniquely identify the nodes based on their 'pre'/'post' values and 'layer'.
        The function automatically checks for the presence of 'pre_activation' and 'post_activation' columns to
        determine whether to color the nodes based on activation values.
        The positions of the nodes are determined by a custom positioning function (`connectome_interpreter.path_finding.create_layered_positions`).
        This function requires the networkx library for graph operations and matplotlib for plotting.
    """
    if path_df.shape[0] == 0:
        raise ValueError("The provided DataFrame is empty.")

    # Create a 'post_layer' column to use as unique identifiers
    if 'post_layer' not in path_df.columns:
        path_df['post_layer'] = path_df['post'].astype(
            str) + '_' + path_df['layer'].astype(str)
    if 'pre_layer' not in path_df.columns:
        path_df['pre_layer'] = path_df['pre'].astype(
            str) + '_' + (path_df.layer-1).astype(str)

    # Create the graph using the new 'post_layer' identifiers
    G = nx.from_pandas_edgelist(path_df, 'pre_layer', 'post_layer', [
                                'weight'], create_using=nx.DiGraph())

    # Labels for nodes
    labels = dict(zip(path_df.post_layer, path_df.post))
    labels.update(dict(zip(path_df.pre_layer, path_df.pre)))

    # Determine the width of the edges
    weights = [G[u][v]['weight'] for u, v in G.edges()]
    widths = [max(0.1, w * 5) for w in weights]  # Scale weights for visibility

    # Generate positions
    from .path_finding import create_layered_positions
    if sort_by_activation:
        node_activation_dict = dict(
            zip(path_df.post_layer, path_df.post_activation))
        node_activation_dict.update(
            dict(zip(path_df.pre_layer, path_df.pre_activation)))
        positions = create_layered_positions(
            path_df, priority_indices, sort_dict=node_activation_dict)
    else:
        positions = create_layered_positions(path_df, priority_indices)

    # Node colors based on activation values
    if ('pre_activation' in path_df.columns) & ('post_activation' in path_df.columns):
        activations = np.concatenate(
            [path_df['pre_activation'].values, path_df['post_activation'].values])
        norm = plt.Normalize(vmin=activations.min(), vmax=activations.max())
        color_map = plt.get_cmap('viridis')
        # Update graph with activation data
        nx.set_node_attributes(
            G, dict(zip(path_df.pre_layer, path_df.pre_activation)), 'activation')
        nx.set_node_attributes(
            G, dict(zip(path_df.post_layer, path_df.post_activation)), 'activation')

        node_colors = [color_map(norm(G.nodes[node]['activation']))
                       for node in G.nodes()]

    # Edge colors based on neuron signs
    default_color = 'lightgrey'  # Default edge color

    # specify edge colour based on pre-neuron sign, if available
    edge_colors = []
    for u, _ in G.edges():
        pre_neuron = labels[u]
        if neuron_to_sign and pre_neuron in neuron_to_sign:
            edge_colors.append(sign_color_map.get(
                neuron_to_sign[pre_neuron], default_color))
        else:
            edge_colors.append(default_color)

    # Plot the graph
    _, ax = plt.subplots(figsize=figsize)
    if ('pre_activation' in path_df.columns) & ('post_activation' in path_df.columns):
        nx.draw(G, pos=positions,
                labels=labels,
                with_labels=True, node_size=100,
                node_color=node_colors,
                arrows=True, arrowstyle='-|>', arrowsize=10,
                font_size=8, width=widths, edge_color=edge_colors, ax=ax)
        plt.colorbar(plt.cm.ScalarMappable(
            norm=norm, cmap=color_map), ax=ax, label='Activation', fraction=fraction, pad=pad)
    else:
        nx.draw(G, pos=positions,
                labels=labels,
                with_labels=True, node_size=100,
                node_color='lightblue', arrows=True, arrowstyle='-|>', arrowsize=10, font_size=8, width=widths, ax=ax, edge_color=edge_colors)

    edge_labels = {(u, v): f'{data["weight"]:.{weight_decimals}f}' for u,
                   v, data in G.edges(data=True)}
    nx.draw_networkx_edge_labels(G, pos=positions, edge_labels=edge_labels,
                                 #  font_color='red'
                                 ax=ax)
    ax.set_ylim(0, 1)
    plt.show()



from typing import Callable, Dict, List, Tuple

import torch
import torch.nn as nn
import numpy as np
import pandas as pd
from scipy.sparse import issparse

from tqdm import tqdm

from .compress_paths import result_summary
from .utils import adjacency_df_to_el, get_activations


class MultilayeredNetwork(nn.Module):
    """
    A PyTorch module representing a multilayered neural network model.
    This network architecture is designed to process temporal sequences of sensory data
    through multiple layers, with the initial layer handling sensory inputs and subsequent
    layers processing sensory + non-sensory input.

    The forward pass of the network unrolls the connectome through time, with each layer
    receiving its own time-specific sensory input.

    Attributes:
        all_weights (torch.nn.Parameter): The connectome. Input neurons are in the columns.
        sensory_indices (list[int]): Indices indicating which rows/columns in the all_weights matrix
            correspond to sensory neurons.
        num_layers (int): The number of layers in the network.
        threshold (float): The activation threshold for neurons in the network.
        activations (numpy.ndarray): A 2D array storing the activations of all neurons (rows) across time steps (columns).

    Args:
        all_weights (torch.Tensor): The connectome. Input neurons are in the columns.
        sensory_indices (list[int]): A list indicating the indices of sensory neurons
            within the network.
        num_layers (int, optional): The number of temporal layers to unroll the network through.
            Defaults to 2.
        threshold (float, optional): The threshold for activation of neurons. Defaults to 0.01.
    """

    def __init__(self, all_weights, sensory_indices, num_layers=2, threshold=0.01, tanh_steepness=5):
        super(MultilayeredNetwork, self).__init__()

        # num_neurons x num_neurons = 18k * 18k, ~1.2GB
        self.all_weights = all_weights  # this does not require grad
        self.sensory_indices = torch.tensor(
            sensory_indices)  # shape: vector of sensory indices. These are the ones we manipulate
        self.num_layers = num_layers
        self.threshold = threshold
        self.tanh_steepness = tanh_steepness
        self.activations = []

    def forward(self, inthroughtime):
        """
        Processes the input through the multilayered neural network, applying thresholding
        and tanh activation functions. This method sequentially processes the input through
        each layer of the network, considering both sensory and non-sensory neurons, to produce
        the final activations.

        Args:
            inthroughtime (torch.Tensor): A 2D tensor (number of neurons, number of time steps) representing the input to the network
                across different time steps. Each column corresponds to a different time step,
                and each row corresponds to different sensory inputs.

        Returns:
            numpy.ndarray: A 2D array of activations for all neurons across all layers and
                time steps. Each column represents the activations for a specific time step,
                while each row corresponds to different neurons.

        This method first applies a threshold ReLU to the input, setting all values below
        the threshold to zero. It then limits the range of activations between 0 and 1 using
        a tanh function. The activation of all neurons is then calculated, based on the input and the weights. The activation of input neurons at layer n is the sum of: input from the previous layer (input neurons are allowed to connect with each other), and freshly fed-in external input.
        The process accounts for the model's threshold and `tanh_steepness` to modulate
        the activations. The final activations are stored in the `activations` attribute
        of the model, with each entry corresponding to the activations for a specific layer.
        """

        self.activations = []  # Clear activations list at each forward pass

        # thresholded relu
        inthroughtime = torch.where(inthroughtime >= self.threshold,
                                    inthroughtime, 0)
        # Limit the range between 0 and 1
        inthroughtime = torch.tanh(inthroughtime)

        # Initial activations are based only on sensory inputs for the first time step
        # shape: (18k , 1)
        x = self.all_weights[:, self.sensory_indices] @ inthroughtime[:, 0]
        # thresholded relu
        x = torch.where(x >= self.threshold, x, 0)

        # Limit the range between 0 and 1
        x = torch.tanh(self.tanh_steepness*x)
        self.activations.append(x)

        # Process remaining layers
        for i in range(1, self.num_layers):
            x = self.all_weights @ x  # shape: (18k, 1)
            # add inthroughtime[:, i] to the sensory neurons
            x[self.sensory_indices] += inthroughtime[:, i]
            # thresholded relu
            x = torch.where(x >= self.threshold, x, 0)

            # Limit the range between 0 and 1
            x = torch.tanh(self.tanh_steepness*x)

            self.activations.append(x)

        # free up memory as much as possible
        del inthroughtime
        del x
        torch.cuda.empty_cache()

        # stack the numpy activations together
        self.activations = torch.stack(self.activations, dim=1)
        return self.activations


def activation_maximisation(
        model, neurons_to_activate: Dict[int, List[int]],
        neurons_to_deactivate=None,
        input_tensor=None,
        num_iterations=100, learning_rate=0.4,
        in_regularisation_lambda=0.1, custom_in_regularisation: Callable[[torch.Tensor], torch.Tensor] = None,
        out_regularisation_lambda=0.1,
        early_stopping=True, stopping_threshold=1e-6, n_runs=10,
        use_tqdm=True, print_output=True, report_memory_usage=False,
        device=None, wandb=True) -> Tuple[np.ndarray, np.ndarray, list, list]:
    """
    Performs activation maximisation on a given model to identify input patterns that maximally activate selected neurons.

    This method adjusts `input_tensor` over `num_iterations` to maximize the activation of specified neurons in `model`,
    while also applying custom regularization to the input and output. The function supports early stopping based on a threshold
    to prevent unnecessary computations if the activation change becomes negligible.

    Args:
        model: A PyTorch model with `activations`, `sensory_indices`, and `threshold` attributes.
        neurons_to_activate (Dict[int, List[int]]): A dictionary mapping from layer indices to lists of neuron indices
            whose activations are to be maximized.
        neurons_to_deactivate (Dict[int, List[int]]): A dictionary mapping from layer indices to lists of neuron indices
            whose activations are to be minimized. Defaults to None.
        input_tensor (torch.Tensor, optional): The initial tensor to optimize. If None, a random tensor is created.
            Defaults to None.
        num_iterations (int, optional): The number of iterations to run the optimization for. Defaults to 100.
        learning_rate (float, optional): The learning rate for the optimizer. Defaults to 0.01.
        in_regularisation_lambda (float, optional): The regularization coefficient for punishing input neuron activation. Defaults to 0.4.
        custom_in_regularisation (Callable[[torch.Tensor], torch.Tensor], optional): A function that applies custom
            regularization to the `input_tensor`. Defaults to L1+L2 norm if None.
        out_regularisation_lambda (float, optional): The coefficient for punishing output neuron activation. Defaults to 0.1.
        early_stopping (bool, optional): Whether to stop the optimization early if the difference between the biggest and the smallest loss within the last n_runs falls below `stopping_threshold`.
            Defaults to True.
        stopping_threshold (float, optional): The threshold for early stopping. Defaults to 1e-6.
        n_runs (int, optional): The number of runs to consider for early stopping. Defaults to 10.
        use_tqdm (bool, optional): Whether to use tqdm progress bars to track optimization progress. Defaults to True.
        print_output (bool, optional): Whether to print loss information during optimization. Defaults to True.
        report_memory_usage (bool, optional): Whether to report GPU memory usage during optimization. Defaults to False.
        device: The device to run the optimization on. If None, automatically selects a device. Defaults to None.
        wandb (bool, optional): Whether to log optimization details to Weights & Biases. Defaults to True. Requires wandb to be installed.

    Returns:
        A tuple containing:
        The optimized `input_tensor` as a numpy array.
        The output of the model after optimization as a numpy array.
        A list of input activation losses over iterations.
        A list of output activation losses over iterations.
        A list of input activation regularization losses over iterations.
        A list of output activation regularization losses over iterations.
        A list of input tensor snapshots taken during optimization.

    Raises:
        ImportError: If `wandb` is True and wandb is not installed.
    """
    # if using Weight & Biases
    if wandb:
        try:
            import wandb
        except ImportError:
            raise ImportError(
                "wandb is not installed. Please install it using 'pip install wandb' or set wandb=False.")

        wandb.init(project="connectome_interpreter")

        wandb.log({'learning_rate': learning_rate,
                  'input_regularization_lambda': in_regularisation_lambda,
                   'output_regularization_lambda': out_regularisation_lambda, })

    if device is None:
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    if input_tensor is None:
        # Make sure this function returns the correct input size for your model
        input_tensor = torch.rand(
            (len(model.sensory_indices), model.num_layers), requires_grad=True, device=device)

    if custom_in_regularisation is None:
        # Define default regularization as L1+L2 if none is provided
        def custom_in_regularisation(x):
            return torch.norm(x, p=1) + torch.norm(x, p=2)

    optimizer = torch.optim.Adam([input_tensor], lr=learning_rate)

    input_snapshots = []
    act_loss = []
    out_reg_loss = []
    in_reg_loss = []
    losses = []
    iteration_range = range(num_iterations)

    if report_memory_usage:
        print('GPU memory before optimization:', torch.cuda.memory_allocated(
            device) / 1e9, 'GB')

    if use_tqdm:
        iteration_range = tqdm(iteration_range)

    for iteration in iteration_range:
        optimizer.zero_grad()
        # Forward pass
        _ = model(input_tensor)

        # Take a snapshot of the input tensor every 5 iterations
        if iteration % 5 == 0:
            # Clone the current state of input_tensor and detach it from the computation graph
            snapshot = input_tensor.clone().detach().cpu().numpy()
            snapshot = np.where(snapshot >= model.threshold,
                                snapshot, 0)  # Thresholded ReLU
            # Limit the range between 0 and 1
            snapshot = np.tanh(snapshot)
            input_snapshots.append(snapshot)

        # Calculate activation loss
        activation_loss = torch.tensor(0.0, device=input_tensor.device)
        for layer_index, neuron_indices in neurons_to_activate.items():
            # Ensure layer index is valid
            if layer_index < model.activations.shape[1]:
                # Get activations for this layer
                layer_activations = model.activations[:, layer_index]
                # Negative sign because we want to maximize activation
                # Only select activations from specified neurons
                activation_loss -= torch.mean(
                    layer_activations[neuron_indices])
            else:
                print(f"Layer index {layer_index} is invalid. Skipping.")
        # in the end, activation loss is the sum of mean activation across layers.

        silence_loss = torch.tensor(0.0, device=input_tensor.device)
        if neurons_to_deactivate is not None:
            for layer_index, neuron_indices in neurons_to_deactivate.items():
                # Ensure layer index is valid
                if layer_index < model.activations.shape[1]:
                    # Get activations for this layer
                    layer_activations = model.activations[:, layer_index]
                    # Only select activations from specified neurons
                    silence_loss += torch.mean(
                        layer_activations[neuron_indices])

        out_regularisation_loss = out_regularisation_lambda * torch.mean(
            model.activations)
        # Apply custom regularisation
        in_regularisation_loss = in_regularisation_lambda * custom_in_regularisation(
            input_tensor)
        loss = activation_loss + silence_loss + \
            in_regularisation_loss + out_regularisation_loss
        losses.append(loss.item())

        if early_stopping and (iteration > n_runs):
            # when the difference between the max and the min < stopping_threshold
            if np.max(losses[-n_runs:]) - np.min(losses[-n_runs:]) < stopping_threshold:
                break

        if wandb:
            dct = {"activation_loss": activation_loss.item(),
                   "silence_loss": silence_loss.item(),
                   "in_regularisation_loss": in_regularisation_loss.item(),
                   'out_regularisation_loss': out_regularisation_loss.item(),
                   "loss": loss.item()}
            wandb.log(dct)

        act_loss.append(activation_loss.item())
        out_reg_loss.append(out_regularisation_loss.item())
        in_reg_loss.append(in_regularisation_loss.item())

        # Backward pass and optimization
        loss.backward()
        optimizer.step()

        # Optional: Print information about the optimization process
        if print_output and iteration % 10 == 0:  # Print every 10 iterations
            print(
                f"Iteration {iteration}: Activation Loss = {activation_loss.item()}, De-activation loss = {silence_loss.item()}, Input regularization Loss = {in_regularisation_loss.item()}, Output regularization Loss = {out_regularisation_loss.item()}")

        if report_memory_usage and iteration % 10 == 0:
            print(f'GPU memory at iteration {iteration}:', torch.cuda.memory_allocated(
                device) / 1e9, 'GB')

        torch.cuda.empty_cache()

    output_after = model(input_tensor).cpu().detach().numpy()
    # first bring to cpu to save gpu memory
    input_tensor = input_tensor.cpu().detach().numpy()
    input_tensor = np.where(input_tensor >= model.threshold,
                            input_tensor, 0)
    # Limit the range between 0 and 1
    input_tensor = np.tanh(input_tensor)

    if report_memory_usage:
        print('GPU memory after optimization:', torch.cuda.memory_allocated(
            device) / 1e9, 'GB')

    return (input_tensor, output_after,
            act_loss, out_reg_loss, in_reg_loss, input_snapshots)


# # implementation where there are no connections between sensory neurons:
# class MultilayeredNetwork(nn.Module):
#     """
#     A PyTorch module representing a multilayered neural network model.
#     This network architecture is designed to process temporal sequences of sensory data
#     through multiple layers, with the initial layer handling sensory inputs and subsequent
#     layers processing sensory + non-sensory input.

#     The forward pass of the network unrolls the connectome through time, with each layer
#     receiving its own time-specific sensory input.

#     Attributes:
#         all_weights (torch.nn.Parameter): The connectome. Input neurons are in the columns.
#         sensory_indices (list[int]): Indices indicating which rows/columns in the all_weights matrix
#             correspond to sensory neurons.
#         num_layers (int): The number of layers in the network.
#         threshold (float): The activation threshold for neurons in the network.
#         activations (list[torch.Tensor]): A list storing the activations of each layer
#             after the forward pass.

#     Args:
#         all_weights (torch.Tensor): The connectome. Input neurons are in the columns.
#         sensory_indices (list[int]): A list indicating the indices of sensory neurons
#             within the network.
#         num_layers (int, optional): The number of temporal layers to unroll the network through.
#             Defaults to 2.
#         threshold (float, optional): The threshold for activation of neurons. Defaults to 0.01.
#     """

#     def __init__(self, all_weights, sensory_indices, num_layers=2, threshold=0.01, tanh_steepness=5):
#         super(MultilayeredNetwork, self).__init__()

#         self.all_weights = torch.nn.parameter.Parameter(all_weights)
#         # Update sensory_indices to reflect the new order
#         # Sensory neurons are now the first ones
#         self.sensory_indices = torch.tensor(sensory_indices)
#         # non_sensory_indices are the rest
#         self.non_sensory_indices = torch.tensor(
#             [i for i in range(all_weights.size(0)) if i not in sensory_indices])
#         self.num_layers = num_layers
#         self.threshold = threshold
#         self.tanh_steepness = tanh_steepness
#         self.activations = []  # List for activations of middle layers

#     def forward(self, inthroughtime):
#         """
#         Processes the input through the multilayered neural network, applying thresholding
#         and tanh activation functions. This method sequentially processes the input through
#         each layer of the network, considering both sensory and non-sensory neurons, to produce
#         the final activations.

#         Args:
#             inthroughtime (torch.Tensor): A 2D tensor representing the input to the network
#                 across different time steps. Each column corresponds to a different time step,
#                 and each row corresponds to different sensory inputs.

#         Returns:
#             torch.Tensor: A 2D tensor of activations for non-sensory neurons across all layers and
#                 time steps. Each column represents the activations for a specific time step,
#                 while each row corresponds to different non-sensory neurons.

#         This method first applies a threshold ReLU to the input, setting all values below
#         the threshold to zero. It then limits the range of activations between 0 and 1 using
#         a tanh function. For each layer (after the initial),the method combines sensory and previous layer's non-sensory activations into a
#         single input tensor, which is then processed to produce the next layer's non-sensory neuron activations.
#         The process accounts for the model's threshold and `tanh_steepness` to modulate
#         the activations. The final activations are stored in the `activations` attribute
#         of the model, with each entry corresponding to the activations for a specific layer.
#         """

#         self.activations = []  # Clear activations list at each forward pass

#         # Ensure the activation of the sensory inputs is less than 1
#         # inthroughtime = torch.sigmoid(inthroughtime)
#         inthroughtime = torch.where(inthroughtime >= self.threshold,
#                                     inthroughtime, 0)  # Thresholded ReLU
#         # Limit the range between 0 and 1
#         inthroughtime = torch.tanh(inthroughtime)

#         # Use broadcasting to create a meshgrid for indexing
#         # The unsqueeze methods are used to align dimensions for broadcasting
#         # .unsqueeze() puts the non_sensory_indices into a column
#         row_indices = self.non_sensory_indices.unsqueeze(
#             1).expand(-1, len(self.sensory_indices))
#         col_indices = self.sensory_indices.unsqueeze(
#             0).expand(len(self.non_sensory_indices), -1)

#         # Initial activations are based only on sensory inputs for the first time step
#         x = self.all_weights[row_indices,
#                              col_indices] @ inthroughtime[:, 0]
#         x = torch.where(x >= self.threshold, x, 0)  # Thresholded ReLU
#         # Limit the range between 0 and 1
#         x = torch.tanh(self.tanh_steepness*x)
#         self.activations.append(x)

#         # Process remaining layers
#         for i in range(1, self.num_layers):
#             # Create an empty tensor to hold both sensory and non-sensory neuron activations.
#             combined_input = torch.zeros(
#                 self.all_weights.shape[0], device=x.device, dtype=x.dtype)
#             # Fill in the sensory inputs at their respective positions.
#             combined_input[self.sensory_indices] = inthroughtime[:, i]
#             # Fill in the non-sensory outputs at their respective positions.
#             combined_input[self.non_sensory_indices] = x

#             x = self.all_weights[self.non_sensory_indices, :] @ combined_input
#             x = torch.where(x >= self.threshold, x, 0)  # Thresholded ReLU
#             # Limit the range between 0 and 1
#             x = torch.tanh(self.tanh_steepness*x)

#             self.activations.append(x)

#         self.activations = torch.stack(self.activations, dim=1)
#         return self.activations


# def activation_maximisation(
#         model, neurons_to_activate: Dict[int, List[int]],
#         input_tensor=None,
#         num_iterations=300, learning_rate=0.4,
#         in_regularisation_lambda=0.1, custom_in_regularisation: Callable[[torch.Tensor], torch.Tensor] = None,
#         out_regularisation_lambda=0.1,
#         early_stopping=True, stopping_threshold=1e-6, n_runs=10,
#         use_tqdm=True, print_output=True,
#         device=None, wandb=True) -> Tuple[np.ndarray, np.ndarray, list, list]:
#     """
#     Performs activation maximisation on a given model to identify input patterns that maximally activate selected neurons.

#     This method adjusts `input_tensor` over `num_iterations` to maximize the activation of specified neurons in `model`,
#     while also applying custom regularization to the input and output. The function supports early stopping based on a threshold
#     to prevent unnecessary computations if the activation change becomes negligible.

#     Args:
#         model: A PyTorch model with `activations`, `sensory_indices`, and `threshold` attributes.
#         neurons_to_activate (Dict[int, List[int]]): A dictionary mapping from layer indices to lists of neuron indices
#             whose activations are to be maximized.
#         input_tensor (torch.Tensor, optional): The initial tensor to optimize. If None, a random tensor is created.
#             Defaults to None.
#         num_iterations (int, optional): The number of iterations to run the optimization for. Defaults to 300.
#         learning_rate (float, optional): The learning rate for the optimizer. Defaults to 0.01.
#         in_regularisation_lambda (float, optional): The regularization coefficient. Defaults to 0.4.
#         custom_in_regularisation (Callable[[torch.Tensor], torch.Tensor], optional): A function that applies custom
#             regularization to the `input_tensor`. Defaults to L1+L2 norm if None.
#         out_regularisation_lambda (float, optional): The coefficient for the output activation loss. Defaults to 0.1.
#         early_stopping (bool, optional): Whether to stop the optimization early if the difference between the biggest and the smallest loss within the last n_runs falls below `stopping_threshold`.
#             Defaults to True.
#         stopping_threshold (float, optional): The threshold for early stopping. Defaults to 1e-6.
#         n_runs (int, optional): The number of runs to consider for early stopping. Defaults to 10.
#         use_tqdm (bool, optional): Whether to use tqdm progress bars to track optimization progress. Defaults to True.
#         print_output (bool, optional): Whether to print loss information during optimization. Defaults to True.
#         device: The device to run the optimization on. If None, automatically selects a device. Defaults to None.
#         wandb (bool, optional): Whether to log optimization details to Weights & Biases. Defaults to True. Requires wandb to be installed.

#     Returns:
#         A tuple containing:
#         The optimized `input_tensor` as a numpy array.
#         The output of the model after optimization as a numpy array.
#         A list of input activation losses over iterations.
#         A list of output activation losses over iterations.
#         A list of regularization losses over iterations.
#         A list of input tensor snapshots taken during optimization.

#     Raises:
#         ImportError: If `wandb` is True and wandb is not installed.
#     """
#     if wandb:
#         try:
#             import wandb
#         except ImportError:
#             raise ImportError(
#                 "wandb is not installed. Please install it using 'pip install wandb' or set wandb=False.")

#         wandb.init(project="connectome_interpreter")

#         wandb.log({'learning_rate': learning_rate,
#                   'input_regularization_lambda': in_regularisation_lambda,
#                    'output_regularization_lambda': out_regularisation_lambda, })
#     if device is None:
#         device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

#     if input_tensor is None:
#         # Make sure this function returns the correct input size for your model
#         input_tensor = torch.rand(
#             (len(model.sensory_indices), model.num_layers), requires_grad=True, device=device)

#     if custom_in_regularisation is None:
#         # Define default regularization as L1+L2 if none is provided
#         def custom_in_regularisation(x):
#             return torch.norm(x, p=1) + torch.norm(x, p=2)

#     optimizer = torch.optim.Adam([input_tensor], lr=learning_rate)

#     input_snapshots = []
#     act_loss = []
#     out_reg_loss = []
#     in_reg_loss = []
#     iteration_range = range(num_iterations)
#     if use_tqdm:
#         iteration_range = tqdm(iteration_range)

#     # map between all indices and non_sensory_only indices
#     # model.non_sensory_indices has indices for the all-to-all matrix
#     all_to_nonsensory_map = {global_idx: num for num,
#                              global_idx in enumerate(model.non_sensory_indices.cpu().detach().numpy())}
#     losses = []
#     for iteration in iteration_range:
#         optimizer.zero_grad()
#         # Forward pass
#         _ = model(input_tensor)

#         if iteration % 5 == 0:
#             # Clone the current state of input_tensor and detach it from the computation graph
#             snapshot = input_tensor.clone().detach().cpu().numpy()
#             snapshot = np.where(snapshot >= model.threshold,
#                                 snapshot, 0)  # Thresholded ReLU
#             # Limit the range between 0 and 1
#             snapshot = np.tanh(snapshot)
#             input_snapshots.append(snapshot)

#         activation_loss = torch.tensor(0.0, device=input_tensor.device)
#         # Check if the model has an 'activations' attribute and the neurons_to_activate is not empty
#         if hasattr(model, 'activations') and neurons_to_activate:
#             for layer_index, neuron_indices in neurons_to_activate.items():
#                 # first transform the neuron_indices to indices of non_sensory neurons only
#                 non_sensory_only_indices = [
#                     all_to_nonsensory_map[idx] for idx in neuron_indices]
#                 # Ensure layer index is valid
#                 if layer_index < model.activations.shape[1]:
#                     # Get activations for this layer
#                     layer_activations = model.activations[:, layer_index]
#                     # Negative sign because we want to maximize activation
#                     # Only select activations from specified neurons
#                     activation_loss -= torch.mean(
#                         layer_activations[non_sensory_only_indices])
#         # in the end, activation loss is the sum of mean activation across layers.

#         out_regularisation_loss = out_regularisation_lambda * torch.mean(
#             model.activations)
#         # Apply custom regularisation
#         in_regularisation_loss = in_regularisation_lambda * custom_in_regularisation(
#             input_tensor)
#         loss = activation_loss + in_regularisation_loss + out_regularisation_loss
#         losses.append(loss.item())

#         if early_stopping and (iteration > n_runs):
#             # when the difference between the max and the min < stopping_threshold
#             if np.max(losses[-n_runs:]) - np.min(losses[-n_runs:]) < stopping_threshold:
#                 break

#         if wandb:
#             dct = {"activation_loss": activation_loss.item(
#             ), "in_regularisation_loss": in_regularisation_loss.item(),
#                 'out_regularisation_loss': out_regularisation_loss.item(),
#                 "loss": loss.item()}
#             wandb.log(dct)

#         act_loss.append(activation_loss.item())
#         out_reg_loss.append(out_regularisation_loss.item())
#         in_reg_loss.append(in_regularisation_loss.item())

#         # Backward pass and optimization
#         loss.backward()
#         optimizer.step()

#         # Optional: Print information about the optimization process
#         if print_output and iteration % 10 == 0:  # Print every 100 iterations
#             print(
#                 f"Iteration {iteration}: Activation Loss = {activation_loss.item()}, Input regularization Loss = {in_regularisation_loss.item()}, Output regularization Loss = {out_regularisation_loss.item()}")

#     output_after = model(input_tensor)
#     input_tensor = torch.where(input_tensor >= model.threshold,
#                                input_tensor, 0)  # Thresholded ReLU
#     # Limit the range between 0 and 1
#     input_tensor = torch.tanh(input_tensor)
#     return (input_tensor.cpu().detach().numpy(),
#             output_after.cpu().detach().numpy(),
#             act_loss, out_reg_loss, in_reg_loss, input_snapshots)


def activations_to_df(inprop, opt_in, out, sensory_indices, inidx_mapping=None, outidx_mapping=None, activation_threshold=0, connectivity_threshold=0):
    """
    Generates a dataframe representing the paths in a layered plot, filtering by activation and connectivity thresholds.

    This function takes the direct connectivity matrix (inprop), optimal input neuron activity, output neuron activity, indices for sensory neurons, and mapping between input and output indices to groups. It generates a dataframe that represents the paths through the network layers.

    Args:
        inprop (scipy.sparse matrix or numpy.ndarray): Matrix representing the synaptic strengths
            between neurons, can be dense or sparse. Presynaptic is in the rows, postsynaptic in the columns.
        opt_in (numpy.ndarray): A 2D array representing optimal input to the network. Neurons are in the rows, timepoints in the columns. Only the first timepoint is used, since `out` is expected to have activity of all neurons, including input neurons.
        out (numpy.ndarray): A 2D array representing the output from the network. The second dimension represents timepoints.
        sensory_indices (list of int): A list of indices corresponding to sensory neurons in `inprop`.
        inidx_mapping (dict, optional): A dictionary mapping indices in `inprop` to new indices. If None, indices are not remapped.
                                       Defaults to None.
        outidx_mapping (dict, optional): A dictionary mapping indices in `out` to new indices. If None, `inidx_mapping` is used for
                                         mapping. Defaults to None.
        activation_threshold (float, optional): A threshold value for activation. Neurons with activations below this threshold are
                                                not considered. Defaults to 0.
        connectivity_threshold (float, optional): A threshold for filtering connections. Connections with weights below this threshold
                                                  are ignored. Defaults to 0.

    Returns:
        pandas.DataFrame: A dataframe representing the paths in the network. Each row is a connection, with columns for 'pre' and
                          'post' neuron indices, 'layer', and their respective activations ('pre_activation', 'post_activation').
    """
    all_indices = list(range(inprop.shape[0]))

    if outidx_mapping is None:
        outidx_mapping = inidx_mapping

    if issparse(inprop):
        inprop = inprop.toarray()

    # get activations from opt_in and out, based on the mappings provided by inidx_mapping and outidx_mapping
    sensory_act = get_activations(
        opt_in, sensory_indices, inidx_mapping, threshold=activation_threshold)
    all_act = get_activations(
        out, all_indices, outidx_mapping, threshold=activation_threshold)

    # get connectivity
    conn = result_summary(inprop, inidx=all_indices,
                          outidx=all_indices, inidx_map=inidx_mapping, outidx_map=outidx_mapping, display_output=False)
    # turn to edgelist, and filter
    conn_el = adjacency_df_to_el(
        conn, threshold=connectivity_threshold)

    # make paths df
    paths = []
    for layer in range(out.shape[1]):
        if layer == 0:
            # the initial layer only has input neuron activations
            pre = sensory_act[layer]
            post = all_act[layer]
        else:
            # all the input, not just the optimised external input
            pre = all_act[layer-1]
            post = all_act[layer]

        connections = conn_el[conn_el.pre.isin(
            pre.keys()) & conn_el.post.isin(post.keys())]
        # so that direct connectivity is layer 1
        connections.loc[:, ['layer']] = layer+1
        connections.loc[:, ['pre_activation']] = connections.pre.map(pre)
        connections.loc[:, ['post_activation']] = connections.post.map(post)
        if connections.shape[0] > 0:
            paths.append(connections)
        else:
            print(f"No connections found in layer {layer+1}.")
    paths = pd.concat(paths)
    return paths



import numpy as np
import pandas as pd

from .utils import to_nparray


def find_path_once(inprop_csc, steps_cpu, inidx, outidx, target_layer_number, top_n=-1, threshold=0):
    """
    Finds the path once between input and output, of distance target_layer_number, returning indices
    of neurons in the previous layer that connect the input with the output. This works by taking the top_n direct upstream partners of the outidx neurons, and intersect those with neurons 'effectively' connected (through steps_cpu) to the inidx neurons.

    Args:
      inprop_csc (scipy.sparse.csc_matrix): The connectivity matrix in Compressed Sparse Column format.
      steps_cpu (list): A list of compressed connectivity matrices: one matrix for each compressed path length.
      inidx (int, float, list, set, numpy.ndarray, or pandas.Series): The input neuron index/indices.
      outidx (int, float, list, set, numpy.ndarray, or pandas.Series): The output neuron index/indices.
      target_layer_number (int): The target layer number to examine. Must be >= 1. When target_layer_number = 1, we are looking at the direct synaptic connectivity.
      top_n (int, optional): The number of top connections to consider based on direct connectivity from inprop_csc. If top_n = -1, all connections are considered.
      threshold (float, optional): The threshold of the direct connectivity from inidx to an average outidx.

    Returns:
      np.ndarray: An array of neuron indices in the previous layer that have significant connectivity, connecting between the `inidx` and `outidx`.
    """

    # make sure they are integers
    inidx = [int(i) for i in inidx]
    outidx = [int(i) for i in outidx]

    inidx = to_nparray(inidx)
    outidx = to_nparray(outidx)

    if target_layer_number == 1:
        # if the target layer is 1, we are looking at the direct synaptic connectivity
        # so we just need to find the indices of the non-zero values in the inprop_csc matrix
        # that correspond to the outidx, and intersect those with the inidx we are interested in.
        colidx = inidx
    else:
        # first get the neurons that effectively connect to inidx, at layer number target_layer_number - 1.
        # for example, in the ORN->PN->KC case, target_layer_number = 2 for the KCs.
        # top_n_row_indices are indices of the PNs that connect to the KCs. So in this case we should use direct connectivity from PNs to get the ORNs.
        # so when target_layer_number == 2, we should use steps_cpu[0] (direct connectivity), that is, target_layer_number-2:
        # subtract 1 for getting top_n_row_indices below which is going one step upstream; and another for 0-based indexing in steps_cpu.
        # the next line gets the targets that receive non-zero compressed input from inidx
        # when target_layer_number >= 2.
        # .nonzero() returns row, column of the nonzero values.
        colidx = steps_cpu[target_layer_number-2][inidx, :].nonzero()[1]

    # then go back one step from the outidx, based on the direct connectivity matrix (inprop)
    us = inprop_csc[:, outidx].nonzero()
    # intersect the non-zero upstream neurons with those effectively connected across layers (colidx)
    # these are still row indices of inprop
    intersect = np.intersect1d(us[0], colidx)

    # direct connectivity between target_layer_number-1 and target_layer_number
    submatrix = inprop_csc[intersect, :][:, outidx]
    # in case outidx is more than one index, calculate the average of each row in the submatrix
    # same length as len(intersect)
    row_averages = np.array(submatrix.mean(axis=1)).flatten()

    # thresholding
    # these are indices of row_averages, which can also be used to index row_averages
    thresholded_indices = np.where(row_averages >= threshold)[0]
    thresholded_row_averages = row_averages[thresholded_indices]
    # thresholded_intersect are a subset of of intersect, which are indices of inprop_csc
    # intersect has the same length as row_averages
    thresholded_intersect = intersect[thresholded_indices]

    # Find the indices (of thresholded_row_averages) of the top n averages
    if top_n == -1:
        top_n_indices = np.argsort(thresholded_row_averages)
    else:
        top_n_indices = np.argsort(thresholded_row_averages)[-top_n:]
    # Get the original row indices corresponding to these top n averages
    top_n_row_indices = thresholded_intersect[top_n_indices]

    # make the edgelist
    submatrix = inprop_csc[top_n_row_indices, :][:, outidx]

    # Convert submatrix to COO format to efficiently find non-zero elements
    coo_submatrix = submatrix.tocoo()

    # Create DataFrame directly from COO format data
    df = pd.DataFrame({
        # Map back to original indices
        'pre': top_n_row_indices[coo_submatrix.row],
        # Map back to original output indices
        'post': outidx[coo_submatrix.col],
        'weight': coo_submatrix.data
    })

    return df


def find_path_iteratively(inprop_csc, steps_cpu, inidx, outidx, target_layer_number, top_n=-1, threshold=0):
    """
    Iteratively finds the path from the specified output (outidx) back to the input (inidx) across
    multiple layers, using the `find_path_once` function to traverse each layer.

    Args:
      inprop_csc (scipy.sparse.csc_matrix): The direct connectivity matrix in Compressed Sparse Column format.
      steps_cpu (np.ndarray): A list of compressed connectivity matrices: one matrix for each compressed path length.
      inidx (int, float, list, set, numpy.ndarray, or pandas.Series): The input neuron indices.
      outidx (int, float, list, set, numpy.ndarray, or pandas.Series): The output neuron indices to start the reverse path finding.
      target_layer_number (int): The number of layers to traverse backwards from the outidx. If target_layer_number = 1, we are looking at the direct synaptic connectivity.
      top_n (int, optional): The number of top connections to consider at each layer based on direct connectivity from inprop_csc. If top_n = -1, all connections are considered.
      threshold (float, optional): The threshold for the average of the direct connectivity from inidx to outidx.

    Returns:
        pd.DataFrame: A DataFrame containing the path data, including the pre-synaptic and post-synaptic neuron indices, the layer (direct connections from inidx: layer = 1), and the weight (input proportion of the postsynaptic neuron) of the direct connection between pre and post.
    """

    inidx = to_nparray(inidx)
    outidx = to_nparray(outidx)

    # path_indices = []  # This will store the path data as a list of arrays
    dfs = []  # This will store the path data as a list of DataFrames

    current_outidx = outidx
    # from target_layer_number to 1, go back one step at a time
    for layer in range(target_layer_number, 0, -1):
        # Find the indices in the current layer that connect to the next layer
        df = find_path_once(
            inprop_csc, steps_cpu, inidx, current_outidx, layer, top_n, threshold)

        # If no indices are found, break the loop as no path can be formed
        if len(df) == 0:
            print(
                'Cannot trace back to the input :(. Try providing a bigger top_n value, or a lower threshold?')
            break

        df['layer'] = layer

        dfs.append(df)

        # Update the outidx for the next iteration to move backwards through layers
        current_outidx = set(df['pre'])

    return pd.concat(dfs)


def create_layered_positions(df, priority_indices=None, sort_dict=None):
    """
    Creates a dictionary of positions for each neuron in the paths, so that the paths can be visualized in a layered manner. It assumes that `df` contains the columns 'layer', 'pre_layer', 'post_layer' (or 'layer', 'pre', 'post'). If a neuron exists in multiple layers, it is plotted multiple times.
    Args:
        df (pd.DataFrame): The DataFrame containing the path data, including the layer number, pre-synaptic index, and post-synaptic index.
        priority_indices (list, set, pd.Series, numpy.ndarray optional): A list of neuron indices that should be plotted on top of each layer. Defaults to None.
        sort_dict (dict, optional): A dictionary of neuron indices as keys and their sorting order as values. Defaults to None.
    Returns:
        dict: A dictionary of positions for each neuron in the paths, with the keys as the neuron indices and the values as the (x, y) coordinates.
    """

    # if post_layer and pre_layer are not present, create them
    if 'post_layer' not in df.columns:
        df['post_layer'] = df['post'].astype(
            str) + '_' + df['layer'].astype(str)
    if 'pre_layer' not in df.columns:
        df['pre_layer'] = df['pre'].astype(
            str) + '_' + (df.layer-1).astype(str)

    if priority_indices is not None:
        priority_indices = set(priority_indices)

    number_of_layers = df.layer.nunique()
    layer_width = 1.0 / (number_of_layers + 1)
    positions = {}

    all_names = set(df['pre_layer']) | set(df['post_layer'])
    name_to_idx = dict(zip(df.pre_layer, df.pre))
    name_to_idx.update(dict(zip(df.post_layer, df.post)))

    for layer in range(0, number_of_layers + 1):
        layer_name = [item for item in all_names if '_'+str(layer) in item]
        if sort_dict is not None:
            layer_name = sorted(layer_name, key=lambda x: sort_dict[x])

        if priority_indices is not None:
            layer_name_priority = [
                item for item in layer_name if name_to_idx[item] in priority_indices]
            layer_name_not = [
                item for item in layer_name if name_to_idx[item] not in priority_indices]

            layer_name = layer_name_not + layer_name_priority
        for index, neuron in enumerate(layer_name, start=1):
            positions[neuron] = (layer * layer_width,
                                 index * 1.0 / (len(layer_name) + 1))

    return positions


def remove_excess_neurons(df, keep=None, target_indices=None, keep_targets_in_middle=False):
    """After filtering, some neurons are no longer on the paths between the input and output neurons. This function removes those neurons from the paths.

    Args:
        df (pd.Dataframe): a filtered dataframe with similar structure as the dataframe returned by `find_path_iteratively()`.
        keep (list, set, pd.Series, numpy.ndarray, optional): A list of neuron indices that should be kept in the paths, even if they don't connect between input and target in the last layer. Defaults to None.
        target_indices (list, set, pd.Series, numpy.ndarray, optional): A list of target neuron indices that should be kept in the last layer. Defaults to None, in which case all neurons in the last layer in `df` would be kept.
        keep_targets_in_middle (bool, optional): If True, the target_indices are kept in the middle layers as well, even if they don't connect between input and target in the last layer. Defaults to False.

    Returns:
        pd.Dataframe: a dataframe with similar structure as the result of `find_path_iteratively()`, with the excess neurons removed.
    """
    if df.layer.max() == 1:
        return df

    # if target_indices are provided, first use this to filter the last layer
    if target_indices is not None:
        target_indices = set(target_indices)
        if not target_indices.issubset(df[df.layer == df.layer.max()].post):
            raise ValueError('The target indices are not in the post-synaptic neurons of the last layer. Here are the indices of the last layer: ',
                             df[df.layer == df.layer.max()].post, '. Your target_indices should be a subset.')

        df = df[(df.layer != df.layer.max()) | df.post.isin(target_indices)]

    while any([set(df[df.layer == i].post) != set(df[df.layer == (i+1)].pre) for i in range(1, df.layer.max())]):
        if keep is not None:
            keep = set(keep)

            if all([set(df[df.layer == i].post).union(keep) == set(df[df.layer == (i+1)].pre).union(keep) for i in range(1, df.layer.max())]):
                break
        else:
            keep = set()

        if keep_targets_in_middle:
            if target_indices is not None:
                keep = keep.union(target_indices)

        # start adding each layer to df_layers_update
        df_layers_update = []
        if df.layer.max() == 2:
            df_layer = df[df.layer == 2]
            df_prev_layer = df[df.layer == 1]

            df_layers_update.append(
                df_layer[df_layer.pre.isin(set(df_prev_layer.post).union(keep))])
            df_layers_update.append(
                df_prev_layer[df_prev_layer.post.isin(set(df_layer.pre).union(keep))])
            df = pd.concat(df_layers_update)

        else:
            for i in range(2, df.layer.max()):
                df_layer = df[df.layer == i]
                df_next_layer = df[df.layer == i+1]
                df_prev_layer = df[df.layer == i-1]

                df_pre = set.intersection(
                    set(df_prev_layer.post) & set(df_layer.pre))
                df_post = set.intersection(
                    set(df_layer.post), set(df_next_layer.pre))

                if i == 2:
                    # add edges in the first layer
                    df_prev_layer = df_prev_layer[df_prev_layer.post.isin(
                        df_pre.union(keep))]
                    df_layers_update.append(df_prev_layer)

                df_layer = df_layer[df_layer.pre.isin(
                    df_pre.union(keep)) & df_layer.post.isin(df_post.union(keep))]
                if df_layer.shape[0] == 0:
                    raise ValueError(
                        'No path found. Try lowering the threshold for the edges to be included in the path.')
                df_layers_update.append(df_layer)

                if i == (df.layer.max()-1):
                    # add edges in the last layer
                    if target_indices is None:
                        df_next_layer = df_next_layer[df_next_layer.pre.isin(
                            df_post.union(keep))]
                    else:
                        df_next_layer = df_next_layer[df_next_layer.pre.isin(
                            df_post.union(keep).union(target_indices))]
                    df_layers_update.append(df_next_layer)

            df = pd.concat(df_layers_update)
    return df


def filter_paths(df, threshold=0, necessary_intermediate=None):
    """Filters the paths based on the weight threshold and the necessary intermediate neurons. The weight threshold refers to the direct connectivity between connected neurons in the path. It is recommended to not put too may neurons in necessary_intermediate, as it may be too stringent and remove all paths.

    Args:
        df (pd.DataFrame): The DataFrame containing the path data, including the layer number, pre-synaptic index, post-synaptic index, and weight.
        threshold (float, optional): The threshold for the weight of the direct connection between pre and post. Defaults to 0.
        necessary_intermediate (dict, optional): A dictionary of necessary intermediate neurons, where the keys are the layer numbers and the values are the neuron indices. Defaults to None.

    Returns:
        pd.DataFrame: The filtered DataFrame containing the path data, including the layer number, pre-synaptic index, post-synaptic index, and weight.
    """
    if threshold > 0:
        df = df[df.weight > threshold]

        df = remove_excess_neurons(df)

    if necessary_intermediate is not None:
        for layer, indices in necessary_intermediate.items():
            if type(indices) != list:
                indices = [indices]
            df = df[df.post.isin(indices) | (df.layer != layer)]

        df = remove_excess_neurons(df)
    return df



from setuptools import setup, find_packages
import pkg_resources

# Function to check if PyTorch is already installed


def is_torch_installed():
    try:
        pkg_resources.get_distribution('torch')
        return True
    except pkg_resources.DistributionNotFound:
        return False


install_requires_list = [
    # List your package dependencies here
    'numpy',
    'pandas',
    'scipy',
    'tqdm',
    'plotly',
    'matplotlib',
    'networkx',
    'seaborn',
    'ipywidgets',
    'IPython',
]
# Add torch only if it's not already installed
if not is_torch_installed():
    install_requires_list.append('torch>=1.7.1')


setup(
    name='connectome_interpreter',
    # If you're making a patch or a minor bug fix, increment the patch version, e.g., from 0.1.0 to 0.1.1.
    # If you're adding functionality in a backwards-compatible manner, increment the minor version, e.g., from 0.1.0 to 0.2.0.
    # If you're making incompatible API changes, increment the major version, e.g., from 0.1.0 to 1.0.0.
    version='1.7.1',
    packages=find_packages(),
    install_requires=install_requires_list,
    extras_require={
        'get_ngl_link': ['nglscenes'],
        'wandb': ['wandb'],
    },
    # Optional metadata
    author='Yijie Yin',
    author_email='yy432@cam.ac.uk',
    description='A tool for connectomics data interpretation',
    keywords=['connectomics', 'neural network',
              'mechanistic interpretability'],
    project_urls={
        'Source': 'https: // github.com/YijieYin/connectome_interpreter',
        'Documentation': 'https://connectome-interpreter.readthedocs.io/en/latest/',
        'Larva connectome example': 'https://colab.research.google.com/drive/1VIMNFBp7dCgN5XOQ9vvzPaqb80BGPZx4?usp=sharing',
        'Adult connectome example (FAFB)': 'https://colab.research.google.com/drive/1ECUagwN-r2rnKyfcYgtR1oG8Lox8m8BW?usp=sharing',
    },
    # url='https://connectome-interpreter.readthedocs.io/en/latest/',
    test_suite='tests'  # Project home page
)
